﻿/** @module digbase
  @mainpage dig - A GUI Library for D
  @section introduction イントロダクション

  dig ライブラリは OS GUI のラッパーとして、今のところ Windows95以降
  のみに対応します。最小限の機能しかないプレα状態と思ってください。
  dig はパブリックドメインですが、net.BurtonRadons.dig.windows.windows
  は例外で LGPL です。


  @section overview 概要

  ダウンロードは @ref downloads ページから利用可能です。
  dig には @ref digc というビルド･ユーティリティがついてきます。

  要約については @ref summary ページをみてください。

  dig へのメインのクラス階層は抽象的なコントロール・クラスから始まります。
  最も重要なものは Frame です。それはウィンドウ・コントロールです。
  レジストリは Registry.singleton の中でカプセルに入れられます。
  一度インストールすれば、"import net.BurtonRadons.dig.main;" を行うことで
  メイン・コントロールを得れます。 "import net.BurtonRadons.dig.gl;" を
  行えば、CanvasGLクラスおよびその多くのコンポーネントを利用できます。


  @section compilation コンパイル

  まずは DMD が必要でしょう。
  <a href="http://www.digitalmars.com/d/dcompiler.html">こちら</a>へ
  行って２つともダウンロードしてください。
  dig は、DMDとDMCがカレントドライブのルート・ディレクトリに
  インストールされ適切に動く状態になっているのを前提にします。

  インストール・ディレクトリ中の "go.bat" プログラムを実行することで
  dig をコンパイルし、インストールします。

  "c:\dmd\bin\..\..\bin" のようなエラーメッセージでおこられるようなら、
  正しくない sc.ini バージョンなので DMD を再インストールしてください。

  同様に、lib.exeからライブラリを用意できないと怒られる場合は、
  dmc.zip の中の正しくないバージョンを上書きするために再び
  dmdalpha.zip を解凍してください。
  (訳注: ファイル名からしてこのへんの事情は古い話の模様)

  インストールするや否や、自動的に例プログラムのうちの1つを
  構築し実行するでしょう。Examples ディレクトリには、
  他にも実行できる例を含んでいます。


  @section usage 使い方

  dig を使うには、一般に以下のような形のファイルを作ります:

  @code
  // digのインポート. ※ CanvasGLやScintillaTextはインポートされない。
  import net.BurtonRadons.dig.main;

  class Program : Frame
  {
      this ()
      {
          caption ("Hello program");

          with (new Label (this))
          {
              grid (0, 0);
              caption ("Hello, world!");
          }
      }
  }

  void main ()
  {
     (new Program).showModal ();
  }
  @endcode

  */

/**

  @page summary Library Summary

  以下は dig のクラス群の要約です。

  @section control コントロール

  コントロールは視覚的な部品です。
  通常トップレベルのコントロールは Frame です。
  コントロールのサブ･クラスとしては以下のようなものがあります:

  <ul>
  <li>#Button        - テキスト･ラベル付の押しボタン。
  <li>#Canvas        - ユーザー描画のコントロール。
  <li>#CanvasGL      - OpenGL コントロール。
  <li>#CheckBox      - %Label 付の on/off ボタン。
  <li>#ComboBox      - 複数のテキスト一覧から選択する。
  <li>#EditText      - 1行あるいは複数行入力のテキストエディタ。
  <li>#Frame         - ウィンドウ。
  <li>#GroupBox      - 1セットのコントロール群を囲むラベル付の境界。
  <li>#Label         - 1行のテキスト表示コントロール。
  <li>#ListBox       - 複数カラムのリスト。
  <li>#Pane          - グリッド合わせ整理。
  <li>#RadioButton   - これらのセットは多数の値の間に選択することを可能にします。
                       #RadioGroup は容易にこのコントロールのセットを
                       管理するために使用することができます。
  <li>#ScintillaText - 極小(Scintilla)テキスト･エディタコントロールへのインターフェース。
  <li>#Spinner       - 数の入力。
  <li>#TreeBox       - 複合のカラムのツリー表示。
  </ul>

  @section dispatchers ディスパッチャ

  イベントの扱いのために、デリゲート一覧のディズパッチャを
  用いてデリゲートを登録します。通常のディスパッチャの型は
  Dispatchar です。
  使用するデリゲートを追加するには以下のようにします:

  @code
  void delegate () func;
  void delegate (Event event) func;
  Dispatcher *reference;
  @endcode

  「イベント」構造は多くのフィールドを持っています。
  １つのイベント型の中でそれらの大部分は使用されません。
  ディスパッチャをリストする時、それらが括弧の中で使用する
  どのフィールドかをタイプしましょう。
  フィールド "flags" は control、lbutton、mbutton、rbutton
  および shift フィールドに言及します。

  下記に control のためのメイン・ディスパッチャを示します。
  まずは一般的なディスパッチャ：

  <ul>
  <li>onDrop      - コントロール上にファイルが落とされた。#DropDispatcher オブジェクトを使用します。
  <li>onHelp      - このコントロールのヘルプが要求された。このデリゲートが空ならば、メッセージはコントロールの親に渡されます。
  <li>onHScroll (scrollType, scrollPoint) - 水平スクロール・バー・アクションが要求された。デリゲートが空ならば、それは水平スクロール・ページによって hscrollPoint を変更し範囲を設定するでしょう。
  <li>onPaint (left, top, right, bottom)  - ペイント要求された。このイベントは Canvas と CanvasGL のサブクラスでのみ発生します。
  <li>onVScroll (scrollType, scrollPoint) - 垂直スクロール・バー・アクションが要求された。デリゲートが空ならば、それは水平スクロール・ページによって vscrollPoint を変更し範囲を設定するでしょう。
  <li>onSizeChanged (x, y)                - コントロールの寸法が変更された。x と y は以前の幅と高さが入っています。
  </ul>

  キーボードに関するディスパッチャは:

  <ul>
  <li>onChar (keyCode, keyRepeat, keyPrevious) - 文字がキーボード上で押さると、このコントロールがフォーカスされる。
  <li>onLostFocus                              - コントロールはもはやキーボードにフォーカスしていない
  </ul>

  マウスに関するディスパッチャは:

  <ul>
  <li>onLButtonDown (x, y, flags) - コントロール上にいるマウスカーソルの、左ボタンが押された。
  <li>onLButtonUp (x, y, flags)   - コントロール上にいるマウスカーソルの、左ボタンが放された。
  <li>onMButtonDown (x, y, flags) - コントロール上にいるマウスカーソルの、中央ボタンが押された。
  <li>onMButtonUp (x, y, flags)   - コントロール上にいるマウスカーソルの、中央ボタンが放された。
  <li>onMouseLeave                - マウスカーソルはコントロール外へ移動した。
  <li>onMouseOver (x, y)          - マウス・カーソルはコントロール上に移動した。
  <li>onMouseWheel (x, y, flags, wheel) - コントロールがキーボードにフォーカスしているおり、マウス・ウィールが回された。
  <li>onRButtonDown               - コントロール上にいるマウスカーソルの、右ボタンが押された。
  <li>onRButtonUp                 - コントロール上にいるマウスカーソルの、右ボタンが放された。
  </ul>


  @section dialog 標準ダイアログ

  共通のオペレーションを行なうための標準ダイアログ・クラスがあります。
　これらのダイアログは次のとおり:

  <ul>
  <li>#ColorSelector - 色選択ダイアログを表示。
  <li>#FileSelector  - ロード/セーブのために１つまたは複数のファイルを選択。
  <li>#FontSelector  - システム・フォントの選択。
  </ul>


  @section bitmap ビットマップ

  ビットマップ・クラスは、キャンバス上の描画に使用できるイメージを
  管理します。BitmapLoad機能を使用して、ファイルからビットマップを
  ロードできます。さらに、ブラシで始まる1セットのクラスを使用して、
  ビットマップ上に描くことができ、それは色を返します。
  そのサブ・クラスは次のとおり:

  <ul>
  <li>#BrushBitmap   - ビットマップからのブラシをサンプリングする。
  <li>#BrushBlend    - Photoshopスタイルのレイヤーを用いて複数のブラシをブレンドする。
  <li>#BrushColor    - 1色を用いる。
  <li>#BrushFilter   - 子ブラシでコンヴォルーション・フィルタを実行。
  <li>#BrushGradient - 線形、または、放射状のグラデーションをサンプリング。
  <li>#BrushNoise    - ランダムに子ブラシの色を変える。
  </ul>

  ブラシおよびビットマップは双方とも回旋できます。
  ブラシに適用された場合は、コンボリューションはサンプル位置を変更します。
  ビットマップに適用された場合は、コンボリューションはシェイプ・ポイント
  を歪めます。
  コンボリューションは次のとおり:

  <ul>
  <li>#RotateContorter    - 1点の周りを指定角度で回転。
  <li>#SwimContorter      - 波パターン中のサンプル位置へ移動。
  <li>#TransformContorter - 位置に変換マトリクスを適用。
  </ul>

  @section math 数学

  OpenGL 2.0以降と同じ役目の名前を持った多くの数学向構造体があります。
  次のとおり:

  <ul>
  <li>#box2    - 2次元 Axis-Aligned Bounding Box。(辺が座標軸に平行な直方体)
  <li>#box3    - 3次元 Axis-Aligned Bounding Box。
  <li>#mat2    - 2x2 マトリックス。
  <li>#mat3    - 3x3 マトリックス。
  <li>#mat4    - 4x4 マトリックス。
  <li>#obb3    - 3次元 有向バウンディングボックス。
  <li>#plane3  - 3次元 ハーフスペース方程式。
  <li>#quat3   - 3次元 回転クウォータニオン。
  <li>#ray3    - 3次元 光線 (ソース+方向)。
  <li>#sphere3 - 3次元 球体。
  <li>#vec2    - 2次元 ベクタ。
  <li>#vec3    - 3次元 ベクタ。
  <li>#vec4    - 4次元 ベクタ。
  </ul>

  */

/**

  @page digc digc

  通常 DMD を用いてコンパイルを行いますが、dig では digc という、
  このプロセスをはるかに簡単にするプログラムが付属します。

  digc は1行で実行ファイルを作るための DMDコンパイラ・ラッパーです。
  プログラム・コマンドの例:

  @code
  digc halhello.d / cartoon.d -then=cartoon -windowed -O -release
  @endcode

   最初に halhello.d をソースとして halhello.exe をビルドします。
  "/" はプログラムを分離します。それらはオプションを共有します。
  halhello.exe のコンパイルに成功すれば、cartoon.d をソースとして
  catoon.exe を作成します。
  "-then" は、全てのコンパイルが成功した後に実行するコマンドを
  記述します。この例では、cartoonデモ が実行されます。

  -windowed は、窓なプログラムをコンパイルするときに指定します。
  指定の無い場合はコンソール・プログラムとしてコンパイルされます。
  WinMain を使う必要はありません。通常の main で適切に動作します。

  -O と -release は DMD の引数です:
  digc はこれらの引数を解析せず、そのまま DMD に渡します。
  使用するライブラリを指定しないことに注意してください。
  digc が代わりに、必要なライブラリを自動的に識別します。
  これについては後述します。

  "-exe=" に続けた名前を実行ファイル名として作成されます。
  例えば、"-exe=halhello".


  @section library ライブラリのコンパイル(作成)

  ライブラリを作成するには、"-lib="に続けて拡張子無しでライブラリ名
  を記述します。１つのライブラリをコンパイルする時、
  digcは、ライブラリ自身のファイル名と同様にコマンドラインで
  与えられたライブラリに注目し、セントラル･レジストリ(/dmd/lib/libraries.dat)
  に記録します。

  .d ファイルが、import されるのと同じ方法でコンパイルされることは重要です。
　例えば、"import package.file;" のように file.d をインポートしたなら、
  コンパイルは "digc file.d" ではなく "digc package/file.d" で行う必要があります。

　実行ファイルにリンクする必要のあるライブラリがどれかを発見するために、
　この情報はコンパイル後に使用されます。

  digc はこのため import 文を求めてファイルを探索し importファイルが
  セントラル・レジストリの中に記録されているかどうかを確かめます。
  そうならば、それはライブラリとその依存性を含んでいます。

  "-shared" オプションを加えることにより共有ライブラリ(WindowsのDLL)を
  作成できます。
  以前にこれをしている人向け: digc は自動的に DllMain を作成し、
  public のシンボルをすべて export します。
  以前にそれをしていない人向け: linuxと同様な振舞いのプログラムになります。
  しかしながら、メイン・プログラムおよび各 DLL は自身の独立したGCを持つでしょう。
  したがって、データはそれらの間で共有することができません。

  */

/**

  @page downloads Downloads and Changelog

  @section v0_0_14 Version 0.0.14

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.14.zip">dig-0.0.14.zip</A>
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.14.zip">dig_html-0.0.14.zip</A>
  - dedit: Tweaks to undo to make it behave more responsively and not take up so much memory.
    It still needs to have an RLE scheme for character differencing.
  - TreeBox: Added.
  - ListBox: More and better semantics.  Rows now have dispatchers that take every translateable message.
  - dedit: Now uses TreeBox for the project view.  This is bound to have introduced bugs.
  - Canvas.textFont can now be used to get the current font.  The default canvas font is made the same as the
    normal font.  I fixed a bug where double-buffering will not get the right font while painting.  Finally, the font
    returned from Canvas.textFont is attached to the canvas, so modifying it causes the canvas text font to be
    changed automatically.
  - Canvas.textFont is REMOVED.  Use Control.font instead.
  - digc now checks the times of the source and dependencies and avoids compilation if files aren't modified.
    This is transitory, since it doesn't check that the compilation arguments are the same.
  - Updating for DMD 0.69.

  @section v0_0_13 Version 0.0.13

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.13.zip">dig-0.0.13.zip</A>
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.13.zip">dig_html-0.0.13.zip</A>
  - dedit: Fixes.  Changing projects no longer crashes inside the GC.
    Project/New puts the new document in the right place.  Saving a new
    document guesses the syntax highlighter.
  - onPaint event is now handed the region that should be painted, and Canvas uses this information to
    do minimal blitting.  You can repaint regions using Control.paintRegion.
  - dedit: Now redraws only the regions which have changed.  For fast computers using normal fonts this
    was a non-issue, but with slow computers or those using ClearType dedit was slow to work with.  It's
    now much faster.
  - dedit: Find (Control-F) has a new drop-down box to search based on syntax highlighting types.
  - dedit: Java syntax highlighter added.
  - digc: Now uses the "module" statement to change the filename to synthesize in recreation.
  - digc: It now passes a page size of 32 to lib.exe, which was bitching when compiling dig.lib with -g on.
  - Checkbox renamed CheckBox to align to the other Boxes.
  - Directory structure change for hygiene and porting concerns.  Use "import net.BurtonRadons.dig.main;" to import dig
    and "import net.BurtonRadons.dig.gl;" to import diggl.
  - Control.SearchList renamed to Control.listPath.
  - Bitmap constructor doesn't require a Control any more.
  - I'm starting to move to UNICODE.  Character strings should be treated as UTF-8 from now on, although
    DMD won't generate them as such yet.
  - Control.focus (control) is no more.  Use control.makeFocus () and control.clearFocus ().
  - All private symbols have been renamed with a digPlatform or digCommon prefix.  There were about five
    different prefix systems in use, it's now standardised.
  - Started the summary page.

  @section v0_0_12 Version 0.0.12

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.12.zip">dig-0.0.12.zip</A>
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.12.zip">dig_html-0.0.12.zip</A>
  - dedit has been made better in many ways.  I am now using Visual Studio 7.0, which cannot be cajoled
    into editing D code properly, and I have been very unimpressed with other editors, so I now use dedit
    entirely.  This means it should be made a better editor very quickly.
  - dedit: Control-Left and Control-Right have been added.  I have been wanting a word-left/-right that was actually
    useful for programming for years now.  I suggest experimenting with it; I think you'll like it much better
    than other editors' handling of this command for macro processing.
  - dedit: Macro recording/playback.
  - You can now install DMD wherever you want so long as you edit the makefile's BASEDIR line to point
    to the location.

  @section v0_0_11 Version 0.0.11

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.11.zip">dig-0.0.11.zip</A>
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.11.zip">dig_html-0.0.11.zip</A>
  - The "-w" option for digc has been changed to "-windowed".
  - libpng is now loaded when needed rather than being linked in.  Faster startup times, less memory use.
  - digc acts a little better now.
  - dstripcopy tool removes unnecessary whitespace from files and function bodies, resulting in faster compilation -
    compiling the cartoon example with the stripped headers is 59% faster than with the full headers.  timerun tool
    just times execution.
  - digc can create DLLs if you supply the -shared option.  It creates the def file (which exports all public symbols,
    like on Linux), generates the import library, and creates a valid DllMain.
  - Brought in dfilter utility, which I use with Doxygen to make D code more suitable for it.
  - Added -install option to digc, which moves the created files into their appropriate locations.
  - When compiling libraries, digc now puts the stripped headers at the end of the .lib file and
    generates those when the library is used, thus removing the need to install anything other
    than the .lib file.
  - Ported over much more of brlib's vector mathematics.  This meant lots more functions for
    vectors, plane3, ray3, box3, sphere3, and quat3.
  - Added image loading support for libjpeg.  Documented ImageLoader a little better.
  - Added image loading support for TGA.  Doesn't support palettised images.
  - Started taking out multiple symbols between digwindows.d and Phobos' windows.d.

  @section v0_0_10 Version 0.0.10

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.10.zip">dig-0.0.10.zip</A>
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.10.zip">dig_html-0.0.10.zip</A>
  - System mouse cursors are now supported in Control.cursor.  The enumeration it takes, Control.Cursor, has a
    diagram showing typical cursors.
  - digc now handles the version statement in determining whether something is really included.
    It also defines the WindowsXP version as appropriate, more to follow.
  - digc creates a resource file containing a manifest so that visual themes are used under Windows XP.
    Custom-drawn controls also use themes and should be backwards portable to previous versions of Windows.
  - Control.onMouseOver is sent when the mouse enters the control's region, paired before the mouseMove.
  - Control.onMouseLeave is sent when the mouse leaves the control's space (Windows 98 and up presently).
  - Timers don't use interrupts anymore, but are instead part of the message feed.
  - Handling the scrollbars is easier now; you only need to use vscrollRange and vscrollPage, and vscrollPoint
    will be updated by onVScroll when it's empty.  The same applies for the horizontal scrollbars.
  - Resizing the window with an empty onResize resizes its contents as appropriate for its stickiness.
  - Event.deltax and .deltay contain the cursor change since the last onMouseMove, or zeroes for the first incidence.
  - Start of a ListBox control.
  - Added a ScintillaText control contributed by Andy Friesen.
  - Updates for DMD 0.57.

  @section v0_0_9 Version 0.0.9

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.9.zip">dig-0.0.9.zip</A>
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.9.zip">dig_html-0.0.9.zip</A>
  - Changes to gridfitting require that you run the "display" method on a control before it's gridfit.  You can do this at any level.
  - digc now records library dependencies in a central registry.

  @section v0_0_8 Version 0.0.8

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.8.zip">dig-0.0.8.zip</A> (317Kb)
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.8.zip">dig_html-0.0.8.zip</A> (238Kb)
  - Took subtypes of GLFrame out into the open.  My only real worries are
    with vec[234]/mat[234], which are OpenGL 2.0-based and could easily
    collide.
  - Enriched vec* and mat* structs with eleven herbs and spices.
    Added plane3, box3, obb3, sphere3.
  - Added menu bars.  Recreation is not allowed presently, so windows with
    menu bars can't have their minimise/maximise/etc. buttons change.
  - Mild changes for DMD 0.49 and 0.50.
  - Event.keyCode is now a string instead of an enumeration.
  - Spinner: Added.  This is a very handy tool.
  - EditText.getFloat now handles expressions.
  - Control.onDrop message for file explorer drag/drop.
  - Renamed TextBox to Label.
  - GroupBox is now user-drawn as it was interfering with some message
    dispatching.
  - GLFrame.readPixels for reading a rectangle of data from one of the
    frame buffers.
  - GLFrame.textureCopyImage and GLtexture.copyImage for copying the
    color buffer into a texture.
  - digc now handles variable expansions so that I could install the dlls.
    It also handles multiple -then statements now.
  - GL functions have been moved into the GL class with a gl singleton, so
    that you don't need to pass around a redundant frame everywhere and can
    step into native symbol lookup.
  - GLFrame was renamed to CanvasGL; a Frame implies a Window.
  - RadioButton control added, as well as the RadioGroup helper.
  - dedit: Find dialog, invoked with Ctrl-F and find next with F3.  Be sure
    to check out the "Match List" button; you click on entries to show that
    point, and can select and delete them.
  - dedit: Horizontal scrollbar.

  @section v0_0_7 Version 0.0.7

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.7.zip">dig-0.0.7.zip</A> (183Kb)
  - Documentation Package: <A href="http://www.opend.org/dig/arc/dig_html-0.0.7.zip">dig_html-0.0.7.zip</A> (172Kb)
  - dedit: Made projects save with relative paths and allowed it to start from the command-line.
  - GroupBox: Fixed to, you know, work.  It should use a manually-drawn widget so that children can be subwindowed directly.
  - Now use a default font as the system default was driving me crazy.  Why this is my responsibility I'll never know.
  - Pane: Added.
  - dedit: Project folders.  Still clumsy.  Cheapo statistics tool.
  - dedit: Fixed copy/paste/undo bugs.
  - digc: Wasn't compiling multiple programs properly.  This was preventing the cartoon demo from being compiled.
  - dedit: Now gives an option to try to save files (to a different filename) after a crash.
  - CheckBox is no longer a subclass of Button.  This should be as flat as possible.

  @section v0_0_6 Version 0.0.6

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.6.zip">dig-0.0.6.zip</A>
  - Added Clipboard class, which has a singleton in Control::clipboard.
    Only supports text content for now.
  - dedit: Added undo and clipboard support.
  - dedit: Backspace is smart to indentation.
  - dedit: HTML syntax highlighting.  Basic as basic can be, but it's contextual.
  - digc now handles wildcards for filenames, used with dedit.
  - digc now builds libraries, used with dig.
  - FileSelector: selecting a single file in multiple-selection mode would return a cancel.
  - dedit: options dialog and session window.
  - Font: Saving to/loading from registry.
  - Registry: Set of "love" methods that either load or save depending upon a bit.
  - dedit: Project manager.

  @section v0_0_5 Version 0.0.5

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.5.zip">dig-0.0.5.zip</A> (159Kb)
  - Removed Registry singleton; it's now a field of Control (Control::registry).
  - ColorWhite, ColorBlack, and ColorTrans are now Color::White, Color::Black, and Color::Trans.
  - Added onClose message and fixed up its behaviour.
  - Better at freeing data.
  - Got rid of Painting, as it was increasingly difficult to see what on Earth it could be used for.
  - Made Canvas use real colors, none of this 0-100 nonsense.
  - Mouse capturing, keyboard lost focus message, #Font.textExtent.
  - Added support for scrollbars.
  - Started dedit example, a sample syntax-highlighting D editor.
    Gave it a bunch of highlighters for some reason.  It's plugin-based,
    to an extent; refactoring will make this clearer.
  - Control::onKeyDown, Control::onChar.
  - gc.fullCollect is now called every second.

  @section v0_0_4 Version 0.0.4 (October 28, 2002)

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.4.zip">dig-0.0.4.zip</A> (124Kb)
  - Added ComboBox control.
  - Deleted windows.d and winutil.d; windows headers are now in digwindows.d, saving 40kb on executable size.
  - Added support for window resizing.  Nigh improveable.
  - Added Registry singleton for registry access.
  - Cartoon shader sample makes use of the registry, combobox, resizing, and has transparent shadows.
  - Added digc and used it for examples.
  - Removed WinMain requirements, digInit and digTerm.
  - Fixed keyboards.

  @section v0_0_3 Version 0.0.3 (October 27, 2002)

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.3.zip">dig-0.0.3.zip</A> (211Kb)
  - Added grid-fitting, that automatically locates controls based on their position in a table (Control::grid, Control::sticky).
  - Added a second example (more a sample), a port of a NeHe cartoon shading tutorial to GLFrame that adds texturing, multiple lights, simple anti-aliased plane shadows, and controls for the parameters.

  @section v0_0_2 Version 0.0.2 (October 26, 2002)

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.2.zip">dig-0.0.2.zip</A> (186Kb)
  - Added Font class (just a blackbox).
  - Added FontSelector.
  - Added ColorSelector.
  - Added GLFrame for OpenGL contexts.
  - Added the start of support for timers (Control::timer).
  - Updated the example program to show these features.

  @section v0_0_1 Version 0.0.1 (October 23, 2002)

  - File: <A href="http://www.opend.org/dig/arc/dig-0.0.1.zip">dig-0.0.1.zip</A> (136Kb)
  - Initial release.

  @section todo TODO List

  - Resizing in cartoon results in mess.
  - Menu has a single dirty scanline.
  - mat3.invert, mat2.invert, vec4's operators.
  - dedit: replace function
    only.  Lowercase and uppercase selection tools.  Show the key binding in the menus,
    key binding setup dialog.
  - dedit: blit context for
    scrolling, line insertion, and line deletion. For scrolling we'll have to buffer and then play
    back before painting, that might work
  - Canvas: Retain the bitmap if it hasn't been resized.
  - dedit: New Project, Close the current file, shouldn't crash.
  - dedit: Closing a modified file
    should produce a warning.  Delete file option.
  - dedit: Keeps locking up.
  - dedit: Two-letter compilation commands act weird.
  - dedit: Saving a project file should add an extension if necessary.
  - dedit: Indent after a local label.
  - dedit: C/C++ auto-indenting.
  - dedit: Return should remove any whitespace after the cursor.
  - dedit: Save All doesn't repaint the project view.
  - dedit: Control-G: Goto line.
  - dedit: Drop-down box for ExternalTool to insert standard expansions.
  - ListBox: Column reordering.  Right-click menu for disabling items.  Content-
    sizing columns.  Partial rendering.  Resizing last column to induce a
    horizontal scrollbar.  Non-resizable columns, non-sortable.  Selections.
  - Use the AlphaBlend function.
  - Functions for drawing path crosses and networks.
  - TrackMouseEvent is stupid, needs some crap.

  */

/**

  @page porting Porting and Standards

  Because dig isn't ported to any other platform, most of this will have
  to be based on speculation, but here's what I have now.

  At its most basic, it must render children in order of insertion.  Children overlapping
  must be handled.  Children that are out-of-range of their parent region must be clipped to
  them.  Painting an overlapped child must not cause flickering.  The exception is
  CanvasGL.  Overlapping a CanvasGL control has an undefined effect.

  Whether the result of a destination is native widgets or simulated doesn't matter to me.
  However, it's unlikely that a native widget will have all of the features needed.

  <ul>
  <li>Platform-specific symbols must be named with a "digPlatform" prefix.

  <li>Common symbols internal to dig must be named with a "digCommon" prefix.  These
  should be used sparingly.

  <li>Classes can have no public fields.  By public I mean those not marked with "digPlatform"
  or "digCommon"; they must not actually be hidden through "protected" or "private".

  <li>Structs can have no private fields.

  <li>American English must be used for symbol names and for documentation.
  I use Proper English normally; I'm biased against American English.  Nevertheless,
  it's the language of too many D users to ignore it.
  </ul>

  */

module net.BurtonRadons.dig.main;

import net.BurtonRadons.dig.common.brush;
import net.BurtonRadons.dig.common.math;
import net.BurtonRadons.dig.common.expressionEvaluator;
import net.BurtonRadons.dig.common.groupBox;
import net.BurtonRadons.dig.common.imageLoader;
import net.BurtonRadons.dig.common.listBox;
import net.BurtonRadons.dig.common.radioGroup;
import net.BurtonRadons.dig.common.treeBox;

import net.BurtonRadons.dig.platform.base;
import net.BurtonRadons.dig.platform.bitmap;
import net.BurtonRadons.dig.platform.button;
import net.BurtonRadons.dig.platform.canvas;
import net.BurtonRadons.dig.platform.checkBox;
import net.BurtonRadons.dig.platform.colorSelector;
import net.BurtonRadons.dig.platform.comboBox;
import net.BurtonRadons.dig.platform.control;
import net.BurtonRadons.dig.platform.editText;
import net.BurtonRadons.dig.platform.fileSelector;
import net.BurtonRadons.dig.platform.font;
import net.BurtonRadons.dig.platform.fontSelector;
import net.BurtonRadons.dig.platform.frame;
import net.BurtonRadons.dig.platform.label;
import net.BurtonRadons.dig.platform.menu;
import net.BurtonRadons.dig.platform.pane;
import net.BurtonRadons.dig.platform.radioButton;
import net.BurtonRadons.dig.platform.spinner;
