// dig ライブラリ内の辻褄合わせように、mbc.d から抜粋したモノ
module net.BurtonRadons.dig.platform.dig_mbs;

version = DIG_MBS;

private import std.string;
private import std.utf;
private import std.c.windows.windows;
import net.BurtonRadons.dig.platform.base;  //DIG_MBS

/// UTF16->MBS. 引数に"文字列"を記述する場合用.
/// toMBS()だと引数の型を決めるためにキャストが必要になるため(本末転倒状態)
char[] _MBS(wchar[] wbuf)
{
    return toMBS(wbuf);
}


/// MBS->UTF16
wchar[] mbsToUTF16(char[] buf)
{
    wchar[] wbuf;
    const int cp = 0;
    wbuf.length = /*wind32.*/MultiByteToWideChar(cp, 0, cast(char*)buf, buf.length, null, 0);
    /*wind32.*/MultiByteToWideChar(cp, 0, cast(char*)buf, buf.length, cast(wchar*)wbuf,wbuf.length);
    return wbuf;
}


/// MBS->UTF8
char[] mbsToUTF8(char[] buf)
{
    return std.utf.toUTF8( mbsToUTF16(buf) );
}


/// MBS->UTF32
dchar[] mbsToUTF32(char[] buf)
{
    return std.utf.toUTF32( mbsToUTF16(buf) );
}


/// UTF16->MBS
char[] toMBS(wchar[] wbuf)
{
    char[] buf;
    const uint cp = 0;
    buf.length = /*wind32.*/WideCharToMultiByte(cp, 0, cast(wchar*)wbuf, wbuf.length, null, 0, null, null);
    /*wind32.*/WideCharToMultiByte(cp, 0, cast(wchar*)wbuf, wbuf.length, cast(char*)buf, buf.length, null, null);
    return buf;
}


/// UTF8->MBS
char[] toMBS(char[] buf)
{
    return toMBS( toUTF16(buf) );
}


/// UTF32->MBS
char[] toMBS(dchar[] dbuf)
{
    return toMBS( toUTF16(dbuf) );
}


/// UTF16 -> MBS (eos='\0')
char* toMBSz(wchar* s)
{
    if (s)
        return std.string.toStringz(toMBS(s[0 .. std.string.wcslen(s)]));
    return "";
}

/// UTF16 -> MBS (eos='\0')
char *toMBSz(wchar[] wbuf)
{
    return  std.string.toStringz( toMBS(wbuf) );
}

/// UTF8 -> MBS (eos='\0')
char* toMBSz(char* s)
{
    if (s)
        return std.string.toStringz(toMBS(s[0 .. std.string.strlen(s)]));
    return "";
}

/// UTF8 -> MBS (eos='\0')
char* toMBSz(char[] buf)
{
    return toMBSz( toUTF16(buf) );
}

/// UTF32 -> MBS (eos='\0')
char* toMBSz(dchar[] dbuf)
{
    return toMBSz( toUTF16(dbuf) );
}


// -------------------------------------------------------------

extern(Windows) export uint GetThreadLocale();

private struct OSVERSIONINFOA {
    DWORD dwOSVersionInfoSize;
    DWORD dwMajorVersion;
    DWORD dwMinorVersion;
    DWORD dwBuildNumber;
    DWORD dwPlatformId;
    CHAR  szCSDVersion[128];
}
private extern(Windows) export BOOL GetVersionExA(OSVERSIONINFOA* lpVersionInformation);


bit  isWinNT()
{
    version (0) {   // all version
        OSVERSIONINFOA osvi;
        memset(&osvi, 0, OSVERSIONINFOA.sizeof);
        osvi.dwOSVersionInfoSize = OSVERSIONINFOA.sizeof;
        bit rc = cast(bit)GetVersionExA(&osvi);
        if (rc != 0)
            rc = osvi.dwPlatformId >= 2/*VER_PLATFORM_WIN32_NT*/;
        return rc;
    } else {
        return 0;
    }
}


bit  isWinXP()
{
    version (0) {   // all version
        OSVERSIONINFOA osvi;
        memset(&osvi, 0, OSVERSIONINFOA.sizeof);
        osvi.dwOSVersionInfoSize = OSVERSIONINFOA.sizeof;
        bit rc = cast(bit)GetVersionExA(&osvi);
        if (rc != 0) {
            rc = osvi.dwPlatformId > 2
                || (osvi.dwPlatformId == 2
                    && (osvi.dwMajorVersion > 5 || (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion >= 1))
                   );
        }
        return rc;
    } else {
        return 0;
    }
}


DWORD GetTabbedTextExtentW_(HDC hDC, LPWSTR lpStr, int nCount, int nTabPos, LPINT pnTabStopPos)
{
    if (isWinNT()) {
        return /*wind32.*/GetTabbedTextExtentW(hDC, lpStr, nCount, nTabPos, pnTabStopPos);
    } else {
        char* mb = toMBSz(lpStr);
        return /*wind32.*/GetTabbedTextExtentA(hDC, mb, strlen(mb)+1, nTabPos, pnTabStopPos);
    }
}

//private extern(Windows) export DWORD GetTabbedTextExtentA(HDC hDC, LPCSTR lpString, int nCount, int nTabPositions, INT *lpnTabStopPositions);
//private extern(Windows) export DWORD GetTabbedTextExtentW(HDC hDC, LPCWSTR lpString, int nCount, int nTabPositions, INT *lpnTabStopPositions);
//extern(Windows) export int MultiByteToWideChar(UINT CodePage, DWORD dwFlags, LPCSTR lpMultiByteStr, int cbMultiByte,LPWSTR lpWideCharStr, int cchWideChar);
//extern(Windows) export int WideCharToMultiByte(UINT CodePage, DWORD dwFlags, LPCWSTR lpWideCharStr, int cchWideChar,LPSTR lpMultiByteStr, int cbMultiByte, LPCSTR lpDefaultChar, LPBOOL  lpUsedDefaultChar);
